//----------------------------------------------------------------------------
//  Body: CReadNI.cpp,v 2.0                                     2010/01/23  //
//                                                                          //
//  Goofie TPC ALICE Program Library                                        //
//  Copyright (C) ALICE collaboration, 2006                                 //
//                                                                          //
//  Definition of class CReadNI                                             //
//  Last modified in CVS                                                    //
//  Author: D. Antonczyk                                                    //
//----------------------------------------------------------------------------

//STL
#include <fstream>
#include <iostream>
#include <sstream>
#include <cmath>

#ifdef ONLINE_ON
  #include "CMacroDef.h"
  #include "COnLineExeption.h"
  #include "CDataStructure.h"
  #include "SConfigData.h"
#endif

#include "CCFileParser.h"
#include "CReadNI.h"

using namespace std;


void CReadAnalogNI::CChannelSettings::SetChannelConfigurations(const uint &_AIChannel, const std::string &_setup)
{
  // set key
  fAI = _AIChannel;
  // what separate parameters
  const string delims("\t,;:|");
  // index
  uint parNo = 0;
  string::size_type begIndex = 0;
  string::size_type endIndex = _setup.find_first_of(delims);
  // look for parameters
  while (endIndex != string::npos) {
    // store values here
    std::string parameter = _setup.substr(begIndex, endIndex-begIndex);
    //
    if(!parameter.empty()) {
      // fill proper variable
      switch(parNo) {
        case 0:
           // first parameter should be physical channel name
           fInfoName = parameter;
        break;

        case 1:
           // channel minimum voltage
           fMinRangeVolts = CReadAnalogNI::StringToNumber<float64>(parameter);
        break;

        case 2:
           // channel maximum voltage
           fMaxRangeVolts = CReadAnalogNI::StringToNumber<float64>(parameter);
        break;

        case 3:
           // channel read mode
           fTerminalConfig = CReadAnalogNI::StringToReadMode(parameter);
        break;

        default:
          cerr << "Warning from <CReadAnalogNI::CChannelSettings::SetMembers>: Too many parameters for one NI channel" << flush << endl;
      }
      // increase counter
      ++parNo;
    }
    // move furhter
    begIndex = endIndex+1;
    endIndex = _setup.find_first_of(delims, endIndex+1);
  }
}
//------------------------ static functions ------------------------------------//

template<typename TR, typename T, typename Trait, typename Alloc>
TR CReadAnalogNI::StringToNumber( const std::basic_string<T, Trait, Alloc>& _toConvert)
{
    std::basic_istringstream< T, Trait, Alloc> ss_toConvert(_toConvert);
    TR retVal;
    ss_toConvert >> retVal;
    return retVal;
}
//-----------------------------------------------------------------------------//

template<typename T, typename Trait, typename Alloc>
int32 CReadAnalogNI::StringToReadMode( const std::basic_string<T, Trait, Alloc>& _toConvert)
{
  int32 retValue = DAQmx_Val_Cfg_Default;
  //
  if(_toConvert.compare("DAQmx_Val_Cfg_Default")==0) return retValue;
  //
  if(_toConvert.compare("DAQmx_Val_RSE")==0) {

    retValue = DAQmx_Val_RSE;
  } else if(_toConvert.compare("DAQmx_Val_Diff")==0) {
     retValue = DAQmx_Val_Diff;

  } else if(_toConvert.compare("DAQmx_Val_NRSE")==0) {
     retValue = DAQmx_Val_NRSE;

  } else {
    cerr << "Warning from <CReadAnalogNI::StringToReadMode>: Channel read mode not specified, using DAQmx_Val_Cfg_Default value" << flush << endl;
  }
  return retValue;
}
//------------------------ constructor -----------------------------------------//

CReadAnalogNI::CReadAnalogNI()
{
   // calibration on
   fCalibrationOnOff     = true;
   // how many channels avaiable
   fTotalChannelsPerCard = 0;
   // default port for NI card installed alone on the system.
   fCardPortUsedForNI    = "Dev1/";
   // errors
   ferror                = 0;                              // NI error code 
   // task
   fTaskHandle           = 0;
   // sample buffor
   for(int v=0; v<10; ++v) {
      fReadArray[v]  = 0;
   }
   //common reading parameters
   fNumSampsPerChan  = 10;                                  // how many samples to read
   fTimeout          = 10.0;                                // in sec
   fFillMode         = DAQmx_Val_GroupByChannel;            // non-interleaved
   fMaxArraySize     = sizeof(fReadArray)/sizeof(float64);  // maximal size of readout array
   fReadArraySize    = 0;	                            // how many sample has been readed
#ifdef ONLINE_ON
   fexeptionStatus   = new COnLineExeption();
#endif
}

//-------------------------- destructor ---------------------------------------//
CReadAnalogNI::~CReadAnalogNI()
{
   // delete configured channels
   fNIChannel.clear();
   // delete channels flag
   fNINotInUseChannel.clear();
   //
#ifdef ONLINE_ON
   if(fexeptionStatus) {
      delete fexeptionStatus;
      fexeptionStatus = NULL;
   }
#endif
}

//--------------------------- initialization -----------------------------------//
uint CReadAnalogNI::InitializeCard(const std::string _configFile)
{
   // Load setup values
   SetNICalValueFromFile(_configFile);
   // Map channels to human readable names
   MapChannelNamesToNumbers();
   // check connection with the NI card.
   double channelRes = ReadChannel(0);
#ifdef ONLINE_ON
   DEBUGINFO("Debug Info => CReadNI::InitializeCard => returned value from channel AI0  = ", channelRes)
#endif
   return ferror;
}

//----------------------------------------------------------------------//
void CReadAnalogNI::SetNICalValueFromFile(const std::string _filename)
{
  try {
    // empty container
    fNIChannel.clear();
    // get maximum number of NI channels
    CCFileParser<MapStringDbl_t> cfg(_filename);
    //
    try {
      fTotalChannelsPerCard = static_cast<int>(cfg.get_value("TotalChannelsPerCard")); // define maximum numbers of channels for given NI card.
    } catch (std::runtime_error _err) {
      //
      cerr << "Warning from <CReadAnalogNI::SetNICalValueFromFile>: Using default number of channels (16)" << flush << endl;
      // set to default value
      fTotalChannelsPerCard = 16;
    }
    // get device/channel setup
    CCFileParser<MapStringString_t> scfg(_filename);
    //
    try {
      fCardPortUsedForNI = scfg.get_value("CardPortUsedForNI");      // set at which port is installed NI card, run 'lsdaq' command to select one.
    } catch (std::runtime_error _err) {
      cerr << "Error in <CReadAnalogNI::SetNICalValueFromFile>: Please set port for NI device properly" << flush << endl;
      throw _err;
    }
    // find configuration for NI channels
    for(int chNo=0; chNo<fTotalChannelsPerCard; ++chNo) {
      //
      try { 
        // set flag that channel is in use
        fNINotInUseChannel[chNo] = false;
	// set channel name	
        stringstream channelID;
        channelID << "AI" << chNo;
       // this function throw runtime error
        std::string channelString = scfg.get_value(channelID.str());
        // set CChannelSettings obj.
        CChannel_t currentChannel;
        //
        currentChannel.first.SetChannelConfigurations(chNo, channelString);
        // get channel calibration parameters
        int parNo = 0;
        // look for parmeters
        while(parNo >= 0) {
          try {
             stringstream channelCalib;
             channelCalib << currentChannel.first.fInfoName << "P" << parNo;
             // this function throw runtime error
             currentChannel.first.fCalibrationPar.push_back(cfg.get_value(channelCalib.str()));
             // next parameter
             ++parNo;
          } catch (std::runtime_error _err) {
            // exit
            parNo = -1;
          }
        }
        // add channel
        fNIChannel[chNo] = currentChannel;
      } catch (std::runtime_error _err) {
        // set flag that channel is in use
        fNINotInUseChannel[chNo] = true;
      }
    }
    //
  } catch(exception& e) {
    string errorInfp = "<CReadNI::SetNICalValueFromFile()>:";
    errorInfp +=  e.what();
#ifdef ONLINE_ON
    fexeptionStatus->SetExeptionError(0x0000001, errorInfp.c_str());
#endif    
  }
}
//----------------------------------------------------------------------//

void CReadAnalogNI::MapChannelNamesToNumbers()
{
   std::map<uint, CChannel_t>::iterator  itNIChannel = fNIChannel.begin();
   //
   for(; itNIChannel != fNIChannel.end(); ++itNIChannel) {
     fMapChannelNameToNumber[itNIChannel->second.first.fInfoName] = itNIChannel->first;
   }
}
//----------------------------------------------------------------------//

double CReadAnalogNI::ReadChannel(const uint _chNr)
{
#ifdef NI_ON
   try {
      // check whether channel has been configured
      if(fNINotInUseChannel[_chNr]) {
         stringstream out;
         out << "Channel " << _chNr << " not configured";
         throw out.str();
      }
      // clean
      fNIChannel[_chNr].second = 0.0;
      // create task
      ferror = DAQmxBaseCreateTask("", &fTaskHandle);
      if(DAQmxFailed(ferror)) throw (int32)-1;
      // create channel
      ferror = DAQmxBaseCreateAIVoltageChan(fTaskHandle,
                                            (fCardPortUsedForNI+fNIChannel[_chNr].first.GetAIChannel()).c_str(),
                                            fNIChannel[_chNr].first.fInfoName.c_str(),
                                            fNIChannel[_chNr].first.fTerminalConfig,
                                            fNIChannel[_chNr].first.fMinRangeVolts,
                                            fNIChannel[_chNr].first.fMaxRangeVolts,
                                            DAQmx_Val_Volts,
                                            NULL);
      if(DAQmxFailed(ferror)) throw (int32)-2;
      // start task
      ferror =  DAQmxBaseStartTask(fTaskHandle);
      if(DAQmxFailed(ferror)) throw (int32)-3;
      // read values
      ferror =  DAQmxBaseReadAnalogF64(fTaskHandle, fNumSampsPerChan, fTimeout, fFillMode, fReadArray, fMaxArraySize, &fReadArraySize, NULL);
      if(DAQmxFailed(ferror)) throw (int32)-4;
   } catch (int32 _error) {
      // error info
      char errBuff[2048]= {'\0'};
      if( DAQmxFailed(_error) ) {
         DAQmxBaseGetExtendedErrorInfo(errBuff, 2048);
         stringstream ss;
         ss << "<CReadNI::ReadChannel()>: " << errBuff;
#ifdef ONLINE_OM         
         fexeptionStatus->SetExeptionError(ferror, ss.str().c_str());
#endif         
      }
   } catch (std::string message) {
     cerr << "Warning from <CReadNI::ReadChannel()>: You are trying to read-out not configured channel, please first configure it. " << message << flush << endl;
     return 0.0;
   }
   // stop/ clear
   if(fTaskHandle != 0) {
      DAQmxBaseStopTask(fTaskHandle);
      DAQmxBaseClearTask(fTaskHandle);
   }
   // calculate average value
   if(fReadArraySize>0) {
     for(int samples=0; samples<fReadArraySize; ++samples) {
       //cout << "test : val ch[" << samples << "] = " << fReadArray[samples] << endl;
       fNIChannel[_chNr].second += (double)fReadArray[samples];
     }
     fNIChannel[_chNr].second /= fReadArraySize;
   } else {
#ifdef ONLINE_ON
     fexeptionStatus->SetExeptionError(ferror, "<CReadNI::ReadChannel()>: read-out of NI card failed");
#endif     
   }
#ifdef ONLINE_ON   
   DEBUGINFO("Debug Info => CReadNI::ReadChannel => read-out from NI card performed correctly", ferror)
#endif
#endif
   return fNIChannel[_chNr].second;
}
//----------------------------------------------------------------------//

double CReadAnalogNI::ReadAndCalibrateChannel(const uint _chNr)
{
  //
  double resValue = 0.0;
  //  read channel
  double measuredValue = ReadChannel(_chNr);
  // chack if channel is read-out correctly
  if(fNINotInUseChannel[_chNr]) return resValue;
  // calibrate
  std::deque<double>::iterator IterCalibrationPar = fNIChannel[_chNr].first.fCalibrationPar.begin();
  // which parameter ?
  int pNo = 0;
  //
  for(; IterCalibrationPar != fNIChannel[_chNr].first.fCalibrationPar.end(); ++IterCalibrationPar) {
    //
    resValue += (*IterCalibrationPar)*std::pow(measuredValue, pNo);
    //
    ++pNo;
  }
  // store calibrated value
  fNIChannel[_chNr].second = resValue;
  //
  return resValue;
}
//----------------------------------------------------------------------//
//-------                        TEST FUNCTIONS                   ------//
//----------------------------------------------------------------------//

void CReadAnalogNI::ShowAcq()
{
   for(int _ch=0; _ch<fTotalChannelsPerCard; ++_ch) {
      if(!fNINotInUseChannel[_ch]) {
	 cout<< " channel[" << _ch << "]=" << fNIChannel[_ch].second << ",\t";
      }
   }
   cout << endl;
}
//----------------------------------------------------------------------//
  
/*!
 *    @param fout is a stream address where data fill be sent.
 */
void CReadAnalogNI::SaveDataInFile(std::ofstream *fout)
{
   //take the data stream and store it in a given file
   if (!fout) {
      cerr << "Error in <CReadAnalogNI::SaveDataInFile(std::ofstream *fout)>: No output file defined " << flush << endl;
   } else {
      for(int _ch=0; _ch<fTotalChannelsPerCard; ++_ch) {
	 if(!fNINotInUseChannel[_ch]) {
	    *fout << " channel[" << _ch << "]=" << fNIChannel[_ch].second << ",\t";
	 }
      }
   }
   *fout << endl;
}
//----------------------------------------------------------------------//
  
/*!
 *  Create a string containing all read-out values.
 *
 *  @return a string with all vlaues read-out from NI channels.
 */
string CReadAnalogNI::GetChannelsValues()
{
   stringstream channelName;
   channelName.str("");

   for(int _ch=0; _ch<fTotalChannelsPerCard; ++_ch) {
      if(!fNINotInUseChannel[_ch]) {
         channelName << " ch[" << _ch << "]=" << fNIChannel[_ch].second << ",\t";
      }
   }
   return channelName.str();
}
//----------------------------------------------------------------------//

//----------------------------------------------------------------------//
//-------        GOOFIE read-out functions                        ------//
//----------------------------------------------------------------------//

CReadNI::CReadNI() : CReadAnalogNI()
{
  ;
}
//----------------------------------------------------------------------//

CReadNI::~CReadNI() 
{
  ;
}
//----------------------------------------------------------------------//

#ifdef ONLINE_ON
unsigned int CReadNI::GetSCData(CDataStructure *_eventStr, bool _update)
{
   if(_update) {
      ReadDataForGOOFIE();
   }
   // set new values
   _eventStr->fSlowControl.sTemp1     = GetChannelValue(fMapChannelNameToNumber["TemperatureSensorNear"]);
   _eventStr->fSlowControl.sTemp2     = GetChannelValue(fMapChannelNameToNumber["TemperatureSensorFar"]);
   _eventStr->fSlowControl.sPress     = GetChannelValue(fMapChannelNameToNumber["PressureSensor"]);
   _eventStr->fSlowControl.sEfVoltage = GetChannelValue(fMapChannelNameToNumber["FieldHV"]);
   _eventStr->fSlowControl.sHVPickup  = GetChannelValue(fMapChannelNameToNumber["PickupVoltage"]);

   return ferror;
}
#endif

//----------------------------------------------------------------------//
void  CReadNI::ReadDataForGOOFIE()
{
   double (CReadAnalogNI::*myPrtReadNI)(const uint) = &CReadAnalogNI::ReadChannel;
   if(fCalibrationOnOff) myPrtReadNI = &CReadAnalogNI::ReadAndCalibrateChannel;

   (this->*myPrtReadNI)(fMapChannelNameToNumber["TemperatureSensorNear"]);
   (this->*myPrtReadNI)(fMapChannelNameToNumber["TemperatureSensorFar"]);
   (this->*myPrtReadNI)(fMapChannelNameToNumber["PressureSensor"]);
   (this->*myPrtReadNI)(fMapChannelNameToNumber["FieldHV"]);
   (this->*myPrtReadNI)(fMapChannelNameToNumber["PickupVoltage"]);
}
