/*!
* @file    CReadNI.h
* @version 2.0
* @author  D. Antonczyk
* @date    2010-23-01
*/
//--------------------------------------------------------------------------//                                                                          //
//  Goofie TPC ALICE Program Library                                        //
//  Copyright (C) ALICE collaboration, 2006                                 //
//--------------------------------------------------------------------------//

#ifndef CREADNI_H
#define CREADNI_H

#include <map>
#include <string>
#include <sstream>
#include <deque>


#ifdef NI_ON
  #include <NIDAQmxBase.h>
  #ifndef ONLINE_ON
    class SConfigData;
    class CDataStructure;
    class COnLineExeption;
  #endif
#else
  #define DAQmx_Val_RSE             1
  #define DAQmx_Val_GroupByChannel  1
  #define DAQmx_Val_NRSE            1
  #define DAQmx_Val_Diff            2
  #define DAQmx_Val_Cfg_Default    -1

  typedef int          int32;
  typedef float        float64;
  typedef unsigned int uInt64;
  typedef bool         bool32;
  typedef int          TaskHandle;
#endif

 typedef unsigned int uint;

/*!
 * @class  CReadAnalogNI
 * @brief  Provide access to analog read-out of the National Instruments M Series borads.
 * @author Dariusz Antonczyk
 */
/*!
 * CReadAnalogNI class in an interface for analog read-out of the National Instruments 
 * M Series Data Acquisition cards. The encapsulated functionality of the interface allow user to configure
 * and read-out individual channels using only few functions call. The configuration in based on
 * a text file which contain following information (example for the goofie system configuration):
 * <pre>
 * //
 * //--------------   NI channels configuration  -----------//
 * //
 * CardPortUsedForNI    = Dev1/                             // set at which port NI card is installed. Run 'lsdaq' command to select correct one.
 * TotalChannelsPerCard = 16                                // define maximum number of channels for a given card.
 * //--------- configure individual channels ---------------//
 * // order of parameters is important and empty characters //
 * // are not allow in the channel_name, i.e.               //
 * // AI# = channel_name; minRange; maxRange; read_mode;    //
 * //-------------------------------------------------------//
 * //
 * AI0 = TemperatureSensorNear; -10.0; 10.0; DAQmx_Val_RSE; // configuration parameters for channel AI0
 * AI1 = TemperatureSensorFar; -10.0; 10.0; DAQmx_Val_RSE;  // configuration parameters for channel AI1
 * AI2 = PressureSensor; -10.0; 10.0; DAQmx_Val_RSE;        // configuration parameters for channel AI2
 * AI3 = FieldHV; -10.0; 10.0; DAQmx_Val_RSE;               // configuration parameters for channel AI3
 * AI4 = PickupVoltage; -10.0; 10.0; DAQmx_Val_RSE;         // configuration parameters for channel AI4
 * //
 * //--------------   NI channels calibration    -----------//
 * // name of calibration parameter must be the same as AI# //
 * // channel_name with prefix P0 for offset and P1 for     //
 * // slope etc ....                                        //
 * // ----------------------------------------------------- //
 * //
 * TemperatureSensorNearP0  = 273.15                        // calibration parameter P0 for goofie temperature sensors near
 * TemperatureSensorNearP1  = 4.0                           // calibration parameter P1 for goofie temperature sensors near
 * TemperatureSensorFarP0   = 273.15                        // calibration parameter P0 for goofie temperature sensors far
 * TemperatureSensorFarP1   = 4.0                           // calibration parameter P1 for goofie temperature sensors far
 * PressureSensorP0         = 825.0                         // calibration parameter P0 for goofie pressure sensors
 * PressureSensorP1         = 37.5                          // calibration parameter P1 for goofie pressure sensors
 * FieldHVP0                = 0.0                           // calibration parameter P0 for goofie field cage voltage (last resistor)
 * FieldHVP1                = 6444.11                       // calibration parameter P1 for goofie field cage voltage (last resistor)
 * PickupVoltageP0          = 0.0                           // not used
 * PickupVoltageP1          = 1.0                           // not used
 * //--------------             END             ------------//
 * </pre>
 *
 * Thus, to add a new read-out channel simplifies to inserting a line into configuration file and mapping
 * it properly in an inherited object, for instance:
 * <pre>
 *   1) In configuration file add a new line:
 *
 *       AI# = MySensor; min_voltage; max_voltage; DAQmx_Val_RSE;
 *
 *      where:
 *            AI#         is a NI channel number '#' used by sensor 'MySensor'. # is from 0 till TotalChannelsPerCard.
 *            MySensor    is a name of the sensor, this name is assigned to read-out value measured by NI card 
 *                           and used in rest of your application. Therefore, chose unique and informative name. 
 *                           In addition 'MySensor' name is used in channel calibration module.
 *            min_voltage is a minimum range, in volts, allowed to be measure by NI channel. See NI manual for more
 *                           information about applied range. Use for instance '-10.0;'.
 *            max_voltage is a maximum range, in volts, allowed to be measure by NI channel. See NI manual for more
 *                           information about applied range. Use for instance '10.0;'.
 *            read_mode      configuring NI channel to read-out in one of the following modes: 'DAQmx_Val_Cfg_Default','
 *                           'DAQmx_Val_RSE', 'DAQmx_Val_Diff', or 'DAQmx_Val_NRSE'.
 *
 *      i.e.: AI5 = TempSensor; -10.0; 10.0; DAQmx_Val_RSE;
 *  
 *      Notice: Each channels parameter must be separate with a one of the punctuation ';' or ',' or '|' or '[TAB]'.
 *              To load configuration file, please call InitializeCard function with file name as an argument.
 *
 *   2) Since CReadAnalogNI is an interface it is necessary to create an inherited object from CReadAnalogNI class 
 *      where user implements mapping into framework. To create a new class inherited from CReadAnalogNI type i.e.:
 *      
 *      class CReadMyDevice : public CReadAnalogNI
 *      {
 *        public:
 *         CReadMyDevice() : CReadAnalogNI() {;}
 *
 *         template<class T>
 *         void GetDataFromChannel(T &_MeasuredValues) {
 *            // check if calibration of channels is required, if so select proper function.
 *            double (CReadAnalogNI::*myPrtReadNI)(const uint) = &CReadAnalogNI::ReadChannel;
 *            if(fCalibrationOnOff) myPrtReadNI = &CReadAnalogNI::ReadAndCalibrateChannel;
 *            // map your object to channel called "TempSensor".
 *            T.TempSensor = (this->*myPrtReadNI)(fMapChannelNameToNumber["TempSensor"]);
 *         }
 *      }
 *
 *      Notice: New channel is mapped via 'MySensor' name which is a configuration parameter, (see point 1).
 *              You are allow to add as many channels to CReadMyDevice class as many your NI card support.
 *
 * </pre>
 *
 * It is also possible to apply a n-th polynomial to calibrate given channel. To introduce calibration, you
 * will need to add into configuration file following entries and set flag to true using SetCalibrationStatus 
 * function, i.e.:
 * <pre>
 * 1) In configuration file add new lines which corresponds to polynomial coefficients. For instance,
 *    if we need to use calibration formula as y = P0 + P1*x + P2*x*x, then add:
 *
 *      MySensorP0 = valueP0
 *      MySensorP1 = valueP1
 *      MySensorP2 = valueP2
 *
 *      Notice: The coefficient name must be the same as NI channel name (MySensor) with prefix P# at the end.
 *              # corresponds to number of polynomial coefficient.
 * </pre>
 */
class CReadAnalogNI
{
 private:
   /*!
    * @struct CChannelSettings
    * @author Dariusz Antonczyk
    * @brief  NI channel settings structure. 
    */
   /*!
     This structure is to keep most frequently configured options for one NI channel.
     It also store calibration parameters for a channel. This structure is a internal
     part of the CReadAnalogNI and it is not foreseen to be use as a stand alone
     object.
    */
   struct CChannelSettings
   {
     int32               fAI;                  //!< Store NI channel name, i.e. (AI0, AI1, etc...)
     std::string         fInfoName;            //!< Store physical channel name, i.e. (TemperatureSensorNear, TemperatureSensorFar, PressureSensor, FieldHV, PickupVoltage)
     float64             fMinRangeVolts;       //!< Set minimum voltage range for channel
     float64             fMaxRangeVolts;       //!< Set maximum voltage range for channel
     int32               fTerminalConfig;      //!< Use one of the following read-out modes (DAQmx_Val_RSE, DAQmx_Val_NRSE, DAQmx_Val_Diff)
     std::deque<double>  fCalibrationPar;      //!< Store calibration parameters, first in deque is P0, then P1, P2, etc.

     /*! @brief Deafult constructor*/
     /*! Set default values for structure members */
     CChannelSettings() : fAI(-1), fMinRangeVolts(-10.0), fMaxRangeVolts(10.0), fTerminalConfig(DAQmx_Val_RSE) {;}

     /*! @brief Class destructor */
     /*! Clear calibration parameters used by channel */
     ~CChannelSettings() {fCalibrationPar.clear();}

     /*! @brief Set structure members from a string contains setup parameters for a channel.*/
     /*! @param _AIChannel a channel number, counted from 0 till fTotalChannelsPerCard.
      *  @param _setup a input string containing all configuration values.
      *
      *  Function is processing _setup string for a given _AIChannel and extracting configuration parameters.
      */
     inline void SetChannelConfigurations(const uint &_AIChannel, const std::string &_setup);

     /*! @brief Function will return AI channel name */
     /*!
      *  @return channel name in format use by NI-DAQmx library.
      */
     std::string GetAIChannel() {std::stringstream out; out << "ai" << fAI; return out.str();}
   };

 public:
   /*! @brief Define type to store channel configuration and measured values.*/
   typedef std::pair<CChannelSettings, double> CChannel_t;

   /*! @brief  Convert string int number. 
    *  @param _toConvert is a string containing some value.
    *  @return given numberic type TR will be return, initialized with value from _toConvert.
    */
   template<typename TR, typename T, typename Trait, typename Alloc>
    static TR StringToNumber( const std::basic_string<T, Trait, Alloc> &_toConvert);
   
   /*! @brief  Convert configuration string into int32 value accepted by NI-DAQmx library.
    *  @param  _toConvert is read mode name.
    */
   template<typename T, typename Trait, typename Alloc>
    static int32 StringToReadMode( const std::basic_string<T, Trait, Alloc> &_toConvert);

   /*! @brief National Instruments M Series Data Acquisition card initialization function.
    *  @param _configFile is a configuration file name. Notice, that full path to file is required
    *                        if you intend to run your code from /bin/ direcotry.
    */
   virtual uint InitializeCard(const std::string _configFile="");

   /*! @param _chNr is a NI channel number, valid range from 0 till fTotalChannelsPerCard.
    *  @return      currently measured value from channel _chNr.
    *  Read-out uncalibrated channel _chNr value. Read-out value will be stored in
    *  CReadAnalogNI::fNIChannel member.
    */
   double       ReadChannel(const uint _chNr);             //!< Read given NI channel.

   /*! @param _chNr is a NI channel number, valid range from 0 till fTotalChannelsPerCard.
    *  @return      currently measured value from channel _chNr.
    *  Read-out calibrated channel _chNr value. Read-out value will be stored in
    *  CReadAnalogNI::fNIChannel member.
    */
   double       ReadAndCalibrateChannel(const uint _chNr); //!< Read given NI channel and calibrate it.

   /*! @brief Get last measured value from a AI channel.
    * @param _chNr is a NI channel number, valid range from 0 till fTotalChannelsPerCard.
    * @return      measured value from channel _chNr. Measure value is read from CReadAnalogNI::fNIChannel object.
    * Function allow user to get more then once read-out value from a given channel. Notice that this
    * function only return stored value for a given channel and not read-out new value from NI card.
    * To update returned value user need to call either  CReadAnalogNI::ReadChannel or 
    * CReadAnalogNI::ReadAndCalibrateChannel function.
    */
   double       GetChannelValue(uint _chNr)           {return fNIChannel[_chNr].second;}

   /*! @brief Get calibration flag status.
    *  @return 'true' or 'false' depending channel calibration is on or off.
    */
   bool         GetCalibrationStatus()                {return fCalibrationOnOff;}

   /*! @brief Set calibration flag status.
    *  @param _cs use to set whether channels calibration shoud be applied. Allow values 'true' or 'false'.
    */
   void         SetCalibrationStatus(bool _cs)        {fCalibrationOnOff = _cs;}

   // Dump measured values
   void         SaveDataInFile(std::ofstream *fout);       //!< Store measered values in an output stream.
   void         ShowAcq();                                 //!< Show measeure values on screan.
   std::string  GetChannelsValues();                       //!< Convert measured value into a string.


protected:
   /*! default constructor */
   CReadAnalogNI();

   /*! class destructor */
   virtual ~CReadAnalogNI();

   /*! Auxiliary function to map channels name readed from configuration file to physical channel number.*/
   void         MapChannelNamesToNumbers();                         //!< Map channels name to NI-AI # number.

   /*! @param _filename is a configuration file name. */
   void         SetNICalValueFromFile(const std::string _filename); //!< Read configuration values from file.

protected:
   /*! 
    *  key is a channel number and value 
    *  corresponds to value stored in Channel_t pair.
    */
   std::map<uint, CChannel_t>  fNIChannel;                //!< Store setup configuartion and measured values for used channels.
   /*!
    *  key is a channel number and value
    *  corresponds to used channel flag.
    */
   std::map<uint, bool>        fNINotInUseChannel;        //!< Store information which channel was used.
   /*!
    *  key is a channel name and value
    *  corresponds to used channel number.
    */
   std::map<std::string, uint> fMapChannelNameToNumber;   //!< Map channel name to channel number. This is a key use in fNIChannel and fNINotInUseChannel maps.
   // apply calibrations
   bool                        fCalibrationOnOff;         //!< Channel calibration on/off.

   int32                       ferror;                    //!< NI error code

#ifdef ONLINE_ON
   /*! error\exception status */
   COnLineExeption            *fexeptionStatus;           //!< Handling error massages.
#endif

private:
   //
   int                         fTotalChannelsPerCard;     //!< How many channels allowed per card/device.
   // 
   std::string                 fCardPortUsedForNI;        //!< Store name of the NI port name, use lsdaq command to obtain it.
   //types defined in NIDAQmxBase
   TaskHandle                  fTaskHandle;               //!< NI task handler
   float64                     fReadArray[10];            //!< Buffer for 10 samples used to make an average of measured values.
   // common reading parameters
   int32                       fNumSampsPerChan;          //!< From how many samples measured value is calculated.
   float64                     fTimeout;                  //!< The amount of time, in seconds, to wait for the function to read the sample(s).
   bool32                      fFillMode;                 //!< Specifies whether or not the samples are interleaved. Allowed values are DAQmx_Val_GroupByChannel - Group by channel (non-interleaved) and DAQmx_Val_GroupByScanNumber - Group by scan number (interleaved).
   uInt64                      fMaxArraySize;             //!< The size of the array, in samples, into which data is read.
   int32                       fReadArraySize;            //!< 
};
//-------------------------------------------------------------------//

/*!
 * @class  CReadNI
 * @brief  Read NI channels used by GOOFIE system.
 * @author Dariusz Antonczyk
 */
/*!
 * Sensor read-out class for GOOFIE system. Mapping sensors with onLine framework.
 *
 */
class CReadNI : public CReadAnalogNI
{
 public:
  CReadNI();
  ~CReadNI();

  /*! 
   *  Read all channels which are relevant for GOOFIE System
   */
  void         ReadDataForGOOFIE();                                     //!< Collect all information needed by onLine analysis framework.

#ifdef ONLINE_ON
  /*! Provide slow control information to the event builder */
  unsigned int GetSCData(CDataStructure *_eventStr, bool _update=TRUE); //!< Function passes read-out values to onLine analysis framework.
#endif
};
//-------------------------------------------------------------------//



/*! @brief Print element of the std contaier.
 *
 * PRINT_ELEMENTS()
 * - prints optional C-string optcstr followed by
 * - all elements of the collection coll
 * - separated by spaces
 */
template <class T>
inline void PRINT_ELEMENTS (const T& coll, const char* optcstr="")
{
    typename T::const_iterator pos;

    std::cout << optcstr;
    for (pos=coll.begin(); pos!=coll.end(); ++pos) {
        std::cout << *pos << ' ';
    }
    std::cout << std::endl;
}
//-------------------------------------------------------------------//

/* Stream function for CReadAnalogNI::CChannelSettings object.
 */
template <typename _Elem, typename _Traits>
std::basic_ostream<_Elem, _Traits> &operator << ( std::basic_ostream<_Elem, _Traits> &_stream, const CReadAnalogNI::CChannelSettings &_ch )
{
   using namespace std;

   ostringstream str;
   str << "********************" << "\n"
       << "CChannelSettings contains:" << "\n";

   str << " int32        fAI             = " << _ch.fAI << endl;
   str << " std::string  fInfoName       = " << _ch.fInfoName << endl;
   str << " float64      fMinRangeVolts  = " << _ch.fMinRangeVolts << endl;
   str << " float64      fMaxRangeVolts  = " << _ch.fMaxRangeVolts << endl;
   str << " int32        fTerminalConfig = " << _ch.fTerminalConfig << endl;
   str << " std::deque<double>           = ";
   _stream << str.str();
   PRINT_ELEMENTS(_ch.fCalibrationPar);
   _stream << endl;

   return _stream;
}

#endif
